#ifndef FOURCC
#define FOURCC(a, b, c, d) ((a)|((b)<<8)|((c)<<16)|((d)<<24))
#endif


#define FCC_ZPAK	FOURCC('Z', 'P', 'A', 'K')


typedef unsigned char byte;
//typedef unsigned short ushort;
//typedef unsigned int uint;
typedef unsigned long long ulong;


struct ZPACK_Header_ds
{
byte magic[4];		//'ZPAK'
byte ents[4];		//size of directory
byte offs[8];		//offset of directory
};

struct ZPACK_DirEntry_ds
{
char name[32];		//file name, 0 padded
byte chain[4];		//next dir entry
byte date_time[4];	//date modified
byte flags[2];		//file flags (1&=fragmented, 2&=dir)
byte method[2];		//method number, 0=store, 8=deflate, 9=deflate(ext)
byte crc32[4];		//CRC32 of data (or first entry if dir)
byte usize[4];		//uncompressed size of data
byte csize[4];		//compressed size of data
byte offset[8];		//offset to start of data
};

struct ZPACK_FragSpan_ds
{
byte _resv[8];		//
byte flags[2];		//fragment flags (1&=dirty)
byte method[2];		//method number, 0=store, 8=deflate, 9=deflate(ext)
byte crc32[4];		//CRC32 of data (or first entry if dir)
byte usize[4];		//uncompressed size of data
byte csize[4];		//compressed size of data
byte offset[8];		//offset to start of data
};

typedef struct
{
ushort flags, method;
uint crc;
uint size;
ulong offs;
}ZPACK_FragSpan;

typedef struct ZPACK_CacheFile_s ZPACK_CacheFile;
typedef struct ZPACK_SpansNode_s ZPACK_SpansNode;

struct ZPACK_CacheFile_s
{
ZPACK_CacheFile *next, *prev;	//mru cache chains
ZPACK_CacheFile *hnext;		//next in current hash
char *path;			//pathname for file
uint lsize, bsize;		//logical and buffer size
uint offs, size;		//offset and size (in image)
byte dirty, method;

byte *buffer;			//cached buffer

ZPACK_FragSpan *frag;		//fragments
int n_frag;			//number of fragments
int m_frag;			//size of fragment array
int frag_bits;			//bits for fragment size
};

struct ZPACK_SpansNode_s
{
int lnode;	//left sub-node
int rnode;	//right sub-node
ulong offs;	//offset/mid-point
uint size;	//size/height
char *name;	//optional name/description
};

typedef struct {
struct ZPACK_DirEntry_ds *root;	//root directory

ZPACK_CacheFile *mru;		//start of cache mru
ZPACK_CacheFile *hash[4096];	//cache hash chains
ZPACK_CacheFile *cachefree;	//free list for cache
ZPACK_CacheFile *cachebuf;	//buffer for cached files

ZPACK_SpansNode *span;		//spans buffer
int spans_root;			//spans root
int spans_free;			//free spans
int n_spans;			//number of spans

FILE *fd;
int offs, ents, ients;		//root offset and ents
uint eof;			//current known end of file
}ZPACK_Context;


typedef struct {
ZPACK_Context *ctx;
ZPACK_CacheFile *cache;
int offs;
}ZPACK_FILE;

typedef struct {
ZPACK_Context *ctx;
int base;	//first entry
int pos;	//current entry
char nbuf[33];	//name buffer
}ZPACK_DIR;

typedef struct {
uint date_time;
ushort flags;
ushort method;
uint crc32;
uint usize;
uint csize;
uint offset;
}ZPACK_STAT;

ZPACK_CacheFile *ZPACK_LookupCache(ZPACK_Context *ctx, char *name);
ZPACK_CacheFile *ZPACK_GetCache(ZPACK_Context *ctx, char *name);
void ZPACK_DropCache(ZPACK_Context *ctx, char *name);
void ZPACK_CommitCacheAll(ZPACK_Context *ctx);

ZPACK_CacheFile *ZPACK_GetCacheFragmented(ZPACK_Context *ctx, char *name);
ZPACK_CacheFile *ZPACK_BufferFragmented(ZPACK_Context *ctx,
	ZPACK_CacheFile *cur);

ZPACK_FILE *ZPACK_OpenFile(ZPACK_Context *ctx, char *name, char *mode);

void ZPACK_InsertSpan(ZPACK_Context *ctx, ulong offs, uint size, char *name);
int ZPACK_DeleteSpan(ZPACK_Context *ctx, ulong offs);
int ZPACK_NextSpan(ZPACK_Context *ctx, ulong offs);
int ZPACK_CheckSpan(ZPACK_Context *ctx, ulong offs, uint size);
ulong ZPACK_FindMinSpan(ZPACK_Context *ctx, uint size);
