ZPACK_CacheFile *ZPACK_GetCacheFragmented(ZPACK_Context *ctx, char *name);
void ZPACK_CommitCacheFragmented(ZPACK_Context *ctx, ZPACK_CacheFile *cur);

void ZPACK_ClearCacheMost(ZPACK_Context *ctx);

ZPACK_CacheFile *ZPACK_AllocCache(ZPACK_Context *ctx)
{
	ZPACK_CacheFile *tmp;
	int i;

	if(!ctx->cachebuf)
	{
		ctx->cachebuf=malloc(1024*sizeof(ZPACK_CacheFile));
		memset(ctx->cachebuf, 0, 1024*sizeof(ZPACK_CacheFile));

		for(i=0; i<1024; i++)
		{
			ctx->cachebuf[i].next=ctx->cachefree;
			ctx->cachefree=&ctx->cachebuf[i];
		}
	}

	if(!ctx->cachefree)
	{
		//no free entries, force files to disk
		return(NULL);	//for now, fail
	}

	tmp=ctx->cachefree;
	ctx->cachefree=tmp->next;

//	memset(tmp, 0, sizeof(ZPACK_CacheFile));
	return(tmp);
}

void ZPACK_FreeCache(ZPACK_Context *ctx, ZPACK_CacheFile *cur)
{
	memset(cur, 0, sizeof(ZPACK_CacheFile));	//sanitize
	cur->next=ctx->cachefree;
	ctx->cachefree=cur;
}


void ZPACK_MoveFrontMRU(ZPACK_Context *ctx, ZPACK_CacheFile *cur)
{
	if(cur->prev)	//relink in mru list
	{
		cur->prev->next=cur->next;
		if(cur->next)cur->next->prev=cur->prev;

		cur->next=ctx->mru;
		if(ctx->mru)ctx->mru->prev=cur;
		ctx->mru=cur;
	}
}

ZPACK_CacheFile *ZPACK_LookupCache(ZPACK_Context *ctx, char *name)
{
	ZPACK_CacheFile *cur;
	char *s;
	int i, j, hi;

	hi=0; s=name;
	while(*s)hi=(hi>>12)^(hi<<3)^(*s++);
	hi&=0xFFF;

	cur=ctx->hash[hi];
	while(cur)
	{
		if(!strcmp(cur->path, name))break;
		cur=cur->hnext;
	}

	return(cur);
}

ZPACK_CacheFile *ZPACK_GetCache(ZPACK_Context *ctx, char *name)
{
	ZPACK_CacheFile *cur;
	byte *buf;
	char *s;
	int i, j, k, hi;

	cur=ZPACK_LookupCache(ctx, name);
	if(cur)
	{
//		printf("Lookup OK: %s\n", name);
		if(cur->frag && !cur->buffer)
			ZPACK_BufferFragmented(ctx, cur);

		ZPACK_MoveFrontMRU(ctx, cur);
		return(cur);
	}

	//ok, file is not cached, so do so

	hi=0; s=name;
	while(*s)hi=(hi>>12)^(hi<<3)^(*s++);
	hi&=0xFFF;

	i=ZPACK_LocateFile(ctx, name);
	if(i<0)	//file does not exist in image
	{
		printf("No File: %s\n", name);
		return(NULL);
	}

	j=ZPACK_GetU16(ctx->root[i].flags);
	if(j&2)
	{
		printf("Cache: Is Directory: %s\n", name);
		return(NULL);	//file is a directory
	}
	if(j&1)
	{
//		printf("Cache: Is Fragmented: %s\n", name);
//		return(NULL);	//for now: file is fragmented

//		printf("Cache: Is Fragmented: %s\n", name);
		cur=ZPACK_GetCacheFragmented(ctx, name);
		if(cur)ZPACK_BufferFragmented(ctx, cur);
		return(cur);
	}

//	cur=malloc(sizeof(ZPACK_CacheFile));
//	memset(cur, 0, sizeof(ZPACK_CacheFile));

	cur=ZPACK_AllocCache(ctx);
	cur->path=strdup(name);

	cur->next=ctx->mru;
	if(ctx->mru)ctx->mru->prev=cur;
	cur->prev=NULL;
	ctx->mru=cur;

	cur->hnext=ctx->hash[hi];
	ctx->hash[hi]=cur;


	cur->offs=ZPACK_GetU64(ctx->root[i].offset);
	cur->size=ZPACK_GetU32(ctx->root[i].csize);
	cur->lsize=ZPACK_GetU32(ctx->root[i].usize);
	cur->method=ZPACK_GetU16(ctx->root[i].method);

	cur->bsize=(cur->lsize+6144)&(~4095);
	cur->buffer=malloc(cur->bsize);

	if(!cur->method)
	{
		fseek(ctx->fd, cur->offs, 0);
		fread(cur->buffer, 1, cur->lsize, ctx->fd);
	}

	if((cur->method==8) || (cur->method==9))
	{
		buf=malloc(cur->size);
		fseek(ctx->fd, cur->offs, 0);
		fread(buf, 1, cur->size, ctx->fd);
		PDUNZ_DecodeStream(buf, cur->buffer,
			cur->size, cur->lsize);
		free(buf);

		j=ZPACK_CRC32(cur->buffer, cur->lsize, (uint)-1);
		k=ZPACK_GetU32(ctx->root[i].crc32);
		if(j!=k)
		{
			printf("ZPACK_GetCache: Decode CRC Problem "
				"(%08X!=%08X)\n", j, k);
			free(cur->buffer);
			cur->buffer=NULL;
		}
	}

	ZPACK_ClearCacheMost(ctx);

	return(cur);
}

void ZPACK_ReleaseCache(ZPACK_Context *ctx, ZPACK_CacheFile *cache)
{
	ZPACK_CacheFile *cur, *lst;
	char *s;
	int i, j, hi;

	hi=0; s=cache->path;
	while(*s)hi=(hi>>12)^(hi<<3)^(*s++);
	hi&=0xFFF;

	cur=ctx->hash[hi]; lst=NULL;
	while(cur)
	{
		if(cur==cache)
		{
			if(cur->prev)	//unlink from mru list
			{
				cur->prev->next=cur->next;
				if(cur->next)cur->next->prev=cur->prev;
			}else
			{
				if(cur->next)cur->next->prev=NULL;
				ctx->mru=cur->next;
			}

			if(lst)lst->hnext=cur->hnext;
				else ctx->hash[hi]=cur->hnext;

			if(cur->buffer)free(cur->buffer);
			if(cur->frag)free(cur->frag);
			free(cur->path);

			ZPACK_FreeCache(ctx, cur);
			break;
		}
		lst=cur; cur=cur->hnext;
	}

	if(!cur)
	{
		printf("ZPACK_ReleaseCache: Cache not present in hash list\n");
	}
}

void ZPACK_DropCache(ZPACK_Context *ctx, char *name)
{
	ZPACK_CacheFile *cur, *lst, *nxt;
	char *s;
	int i, j, hi;

	hi=0; s=name;
	while(*s)hi=(hi>>12)^(hi<<3)^(*s++);
	hi&=0xFFF;

	cur=ctx->hash[hi]; lst=NULL;
	while(cur)
	{
		nxt=cur->hnext;

		if(!strcmp(cur->path, name))
		{
			if(cur->prev)	//unlink from mru list
			{
				cur->prev->next=cur->next;
				if(cur->next)cur->next->prev=cur->prev;
			}else
			{
				if(cur->next)cur->next->prev=NULL;
				ctx->mru=cur->next;
			}

			if(lst)lst->hnext=cur->hnext;
				else ctx->hash[hi]=cur->hnext;

			if(cur->buffer)free(cur->buffer);
			if(cur->frag)free(cur->frag);
			free(cur->path);

//			free(cur);
			ZPACK_FreeCache(ctx, cur);
		}

		lst=cur; cur=nxt;
	}

	return;
}


void ZPACK_CommitCache(ZPACK_Context *ctx, ZPACK_CacheFile *cur)
{
	byte *buf;
	int i, j, sz;

	if(!cur->dirty)return;

	if(cur->frag)
	{
		ZPACK_CommitCacheFragmented(ctx, cur);
		return;
	}

	cur->method=0;

	buf=malloc(2*cur->lsize);
	sz=PDZ2_EncodeStream(cur->buffer, buf, cur->lsize, 2*cur->lsize);
	if((sz>=0) && (sz<(cur->lsize*0.9)))
		cur->method=8;

	if(!cur->method)
	{
		free(buf);
		buf=cur->buffer;
		sz=cur->lsize;
	}

//	if(sz>cur->size)
	if(!ZPACK_CheckSpan(ctx, cur->offs, sz))
	{
//		cur->offs=ctx->eof;
//		ctx->eof+=sz;

		ZPACK_DeleteSpan(ctx, cur->offs);

		cur->offs=ZPACK_FindMinSpan(ctx, sz);
		if((cur->offs+sz)>ctx->eof)ctx->eof=cur->offs+sz;
	}
	cur->size=sz;

	//make sure span is up to date
	ZPACK_InsertSpan(ctx, cur->offs, cur->size, cur->path);

	fseek(ctx->fd, 0, 2);
	i=ftell(ctx->fd);
	if(i<cur->offs)ZPACK_WritePad(ctx->fd, i-cur->offs);
	fseek(ctx->fd, cur->offs, 0);
	fwrite(buf, 1, cur->size, ctx->fd);

	cur->dirty=0;

	if(cur->method)free(buf);


	i=ZPACK_LocateFile(ctx, cur->path);
	if(i<0)
	{
		printf("Commit Cache: can't locate %s\n", cur->path);
		return;
	}

	j=ZPACK_CRC32(cur->buffer, cur->lsize, (uint)-1);

	ZPACK_SetU64(ctx->root[i].offset, cur->offs);
	ZPACK_SetU32(ctx->root[i].csize, cur->size);
	ZPACK_SetU32(ctx->root[i].usize, cur->lsize);
	ZPACK_SetU16(ctx->root[i].method, cur->method);

	ZPACK_SetU32(ctx->root[i].crc32, j);
}

void ZPACK_CommitCacheAll(ZPACK_Context *ctx)
{
	ZPACK_CacheFile *cur;

	cur=ctx->mru;
	while(cur)
	{
		ZPACK_CommitCache(ctx, cur);
		cur=cur->next;
	}
}

void ZPACK_ClearCacheAll(ZPACK_Context *ctx)
{
	ZPACK_CacheFile *cur, *nxt;

	cur=ctx->mru;
	while(cur)
	{
		nxt=cur->next;
		ZPACK_CommitCache(ctx, cur);
		ZPACK_ReleaseCache(ctx, cur);
		cur=nxt;
	}
}

void ZPACK_ClearCacheMost(ZPACK_Context *ctx)
{
	ZPACK_CacheFile *cur, *nxt;
	int i, j;

	i=0; j=0;
	cur=ctx->mru;
	while(cur)
	{
		i++; j+=cur->lsize;

		if((i<8) || ((i<512) && (j<(1<<24))))
		{
			cur=cur->next;
			continue;
		}

		nxt=cur->next;
		ZPACK_CommitCache(ctx, cur);
		ZPACK_ReleaseCache(ctx, cur);
		cur=nxt;
	}
}


ZPACK_CacheFile *ZPACK_GetCacheFragmented(ZPACK_Context *ctx, char *name)
{
	struct ZPACK_FragSpan_ds *span;
	ZPACK_CacheFile *cur;
	byte *buf;
	char *s;
	int i, j, k, hi, lsz, bits;

	cur=ZPACK_LookupCache(ctx, name);
	if(cur)
	{
//		printf("Lookup OK: %s\n", name);
		ZPACK_MoveFrontMRU(ctx, cur);
		return(cur);
	}

	//ok, file is not cached, so do so

	hi=0; s=name;
	while(*s)hi=(hi>>12)^(hi<<3)^(*s++);
	hi&=0xFFF;

	i=ZPACK_LocateFile(ctx, name);
	if(i<0)	//file does not exist in image
	{
		printf("No File: %s\n", name);
		return(NULL);
	}

	cur=ZPACK_AllocCache(ctx);
	cur->path=strdup(name);

	cur->next=ctx->mru;
	if(ctx->mru)ctx->mru->prev=cur;
	cur->prev=NULL;
	ctx->mru=cur;
	cur->hnext=ctx->hash[hi];
	ctx->hash[hi]=cur;


	cur->offs=ZPACK_GetU64(ctx->root[i].offset);
	cur->size=ZPACK_GetU32(ctx->root[i].csize);
	lsz=ZPACK_GetU32(ctx->root[i].usize);
	cur->method=ZPACK_GetU16(ctx->root[i].method);

	span=malloc(lsz);
	if(!cur->method)
	{
		fseek(ctx->fd, cur->offs, 0);
		fread(span, 1, lsz, ctx->fd);
	}

	if((cur->method==8) || (cur->method==9))
	{
		buf=malloc(cur->size);
		fseek(ctx->fd, cur->offs, 0);
		fread(buf, 1, cur->size, ctx->fd);
		PDUNZ_DecodeStream(buf, (byte *)span, cur->size, lsz);
		free(buf);

		j=ZPACK_CRC32(span, lsz, (uint)-1);
		k=ZPACK_GetU32(ctx->root[i].crc32);
		if(j!=k)
		{
			printf("ZPACK_GetCacheFragmented: Decode CRC Problem "
				"(%08X!=%08X)\n", j, k);
			free(span);
			cur->buffer=NULL;

			return(NULL);
		}
	}

	cur->n_frag=lsz/sizeof(struct ZPACK_FragSpan_ds);
	cur->m_frag=cur->n_frag+(cur->n_frag>>1);
	cur->frag=malloc(cur->m_frag*sizeof(ZPACK_FragSpan));
	memset(cur->frag, 0, cur->m_frag*sizeof(ZPACK_FragSpan));

	j=ZPACK_GetU32(span[0].usize);
	i=0;
	while(j>1)
	{
		if(j&1)
		{
			printf("ZPACK_GetCacheFragmented: "
				"Non-Log2 frag size\n");
			return(NULL);
		}
		i++; j>>=1;
	}
	cur->frag_bits=i;
	bits=cur->frag_bits;

//	printf("Cache Frag bits %d\n", i);

	j=0;
	for(i=0; i<cur->n_frag; i++)
	{
		cur->frag[i].flags=ZPACK_GetU16(span[i].flags);
		cur->frag[i].method=ZPACK_GetU16(span[i].method);
		cur->frag[i].crc=ZPACK_GetU32(span[i].crc32);
		cur->frag[i].size=ZPACK_GetU32(span[i].csize);
		cur->frag[i].offs=ZPACK_GetU32(span[i].offset);

		ZPACK_InsertSpan(ctx,
			cur->frag[i].offs, cur->frag[i].size,
			cur->path);

		j+=ZPACK_GetU32(span[i].usize);
	}

//	printf("Cache Frag File Size %d\n", j);
	cur->lsize=j;

	ZPACK_ClearCacheMost(ctx);
	return(cur);
}

ZPACK_CacheFile *ZPACK_BufferFragmented(ZPACK_Context *ctx,
	ZPACK_CacheFile *cur)
{
	byte *buf;
	int i, j, k, bits;

	bits=cur->frag_bits;

//	cur->bsize=(cur->lsize+6144)&(~4095);
	cur->bsize=cur->n_frag<<bits;
	cur->buffer=malloc(cur->bsize);
	memset(cur->buffer, 0, cur->bsize);

	for(i=0; i<cur->n_frag; i++)
	{
//		printf("R-Frag: %d %08X..%08X (%d)\n", i,
//			(uint)cur->frag[i].offs,
//			(uint)cur->frag[i].offs+cur->frag[i].size,
//			cur->frag[i].size);

		fseek(ctx->fd, cur->frag[i].offs, 0);
		if(cur->frag[i].method==0)
		{
			fread(cur->buffer+(i<<bits),
				1, cur->frag[i].size, ctx->fd);
			continue;
		}
		if((cur->frag[i].method==8) || (cur->frag[i].method==9))
		{
			buf=malloc(cur->frag[i].size);
			fread(buf, 1, cur->frag[i].size, ctx->fd);
			PDUNZ_DecodeStream(buf, cur->buffer+(i<<bits),
				cur->frag[i].size, 1<<bits);
			free(buf);
			continue;
		}

		k=1<<bits;
		if(i=(cur->n_frag-1))k=cur->lsize&((1<<bits)-1);
		j=ZPACK_CRC32(cur->buffer+(i<<bits), k, (uint)-1);

		if((uint)j!=cur->frag[i].crc)
		{
			printf("%08X!=%08X\n", j, cur->frag[i].crc);
			free(cur->buffer);
			cur->buffer=NULL;
			return(cur);
		}
	}
}

void ZPACK_CommitCacheFragment(ZPACK_Context *ctx,
	ZPACK_CacheFile *cur, int num)
{
	byte *buf;
	int i, j, sz, bits;

	if(!(cur->frag[num].flags&1))return;

	bits=cur->frag_bits;

	j=ZPACK_CRC32(cur->buffer+(num<<bits), 1<<bits, (uint)-1);
	if(cur->frag[num].crc==j)
	{
		//data is, theoretically, unchanged
		cur->frag[num].flags&=~1;
		return;
	}
	cur->frag[num].crc=j;
	cur->dirty=1;


	cur->frag[num].method=0;

	i=1<<bits;
	if(num==(cur->n_frag-1))i=cur->lsize&((1<<bits)-1);

	buf=malloc(2<<bits);
	sz=PDZ2_EncodeStream(cur->buffer+(num<<bits), buf, i, 2<<bits);
	if((sz>=0) && (sz<((1<<bits)*0.9)))
		cur->frag[num].method=8;

	if(!cur->frag[num].method)
	{
		free(buf);
		buf=cur->buffer+(num<<bits);
		sz=1<<bits;
	}

	if(!ZPACK_CheckSpan(ctx, cur->frag[num].offs, sz))
	{
		ZPACK_DeleteSpan(ctx, cur->frag[num].offs);
		cur->frag[num].offs=ZPACK_FindMinSpan(ctx, sz);
		if((cur->frag[num].offs+sz)>ctx->eof)
			ctx->eof=cur->frag[num].offs+sz;
	}
	cur->frag[num].size=sz;

	//make sure span is up to date
	ZPACK_InsertSpan(ctx, cur->frag[num].offs,
		cur->frag[num].size, cur->path);

	printf("Frag: %d %08X..%08X (%d) %d\n",
		num, (uint)cur->frag[num].offs,
		(uint)cur->frag[num].offs+cur->frag[num].size,
		cur->frag[num].size,
		cur->frag[num].method);

	fseek(ctx->fd, 0, 2);
	i=ftell(ctx->fd);
	if(i<cur->frag[num].offs)
		ZPACK_WritePad(ctx->fd, i-cur->frag[num].offs);
	fseek(ctx->fd, cur->frag[num].offs, 0);
	fwrite(buf, 1, cur->frag[num].size, ctx->fd);

	if(!cur->frag[num].method)free(buf);

	cur->frag[num].flags&=~1;
}

void ZPACK_CommitCacheFragmented(ZPACK_Context *ctx, ZPACK_CacheFile *cur)
{
	struct ZPACK_FragSpan_ds *span;
	byte *buf;
	int i, j, k, sz, lsz;

	cur->dirty=0;
	for(i=0; i<cur->n_frag; i++)
		ZPACK_CommitCacheFragment(ctx, cur, i);
	if(!cur->dirty)return;	//no fragments changed

	lsz=cur->n_frag*sizeof(struct ZPACK_FragSpan_ds);
	span=malloc(lsz);
	memset(span, 0, lsz);

	for(i=0; i<cur->n_frag; i++)
	{
		j=1<<cur->frag_bits;
		if(i==(cur->n_frag-1))
			j=cur->lsize&((1<<cur->frag_bits)-1);

		ZPACK_SetU16(span[i].flags, cur->frag[i].flags);
		ZPACK_SetU16(span[i].method, cur->frag[i].method);
		ZPACK_SetU32(span[i].crc32, cur->frag[i].crc);
		ZPACK_SetU32(span[i].usize, j);
		ZPACK_SetU32(span[i].csize, cur->frag[i].size);
		ZPACK_SetU64(span[i].offset, cur->frag[i].offs);
	}


	buf=malloc(2*lsz);
	sz=PDZ2_EncodeStream((byte *)span, buf, lsz, 2*lsz);
	cur->method=8;

	if(!ZPACK_CheckSpan(ctx, cur->offs, sz))
	{
		ZPACK_DeleteSpan(ctx, cur->offs);
		cur->offs=ZPACK_FindMinSpan(ctx, sz);
		if((cur->offs+sz)>ctx->eof)ctx->eof=cur->offs+sz;
	}
	cur->size=sz;

	//make sure span is up to date
	ZPACK_InsertSpan(ctx, cur->offs, cur->size, cur->path);

	fseek(ctx->fd, 0, 2);
	i=ftell(ctx->fd);
	if(i<cur->offs)ZPACK_WritePad(ctx->fd, i-cur->offs);
	fseek(ctx->fd, cur->offs, 0);
	fwrite(buf, 1, cur->size, ctx->fd);

	cur->dirty=0;

	i=ZPACK_LocateFile(ctx, cur->path);
	if(i<0)
	{
		printf("Commit Cache (Frag): can't locate %s\n", cur->path);
		return;
	}

	j=ZPACK_GetU16(ctx->root[i].flags);
	ZPACK_SetU16(ctx->root[i].flags, j|1);

	j=ZPACK_CRC32(span, lsz, (uint)-1);

	ZPACK_SetU64(ctx->root[i].offset, cur->offs);
	ZPACK_SetU32(ctx->root[i].csize, cur->size);
	ZPACK_SetU32(ctx->root[i].usize, lsz);
	ZPACK_SetU16(ctx->root[i].method, cur->method);

	ZPACK_SetU32(ctx->root[i].crc32, j);

	free(buf);
	free(span);
}
