void ZPACK_CreateImage(char *name)
{
	struct ZPACK_Header_ds hdr;
	FILE *fd;

	ZPACK_SetU32(hdr.magic, FCC_ZPAK);
	ZPACK_SetU32(hdr.ents, 63);
	ZPACK_SetU64(hdr.offs, 64);

	fd=fopen(name, "wb");
	fwrite(&hdr, 1, 16, fd);
	ZPACK_WritePad(fd, 4096-16);
	fclose(fd);
}

void ZPACK_ProcessDir_R(ZPACK_Context *ctx, int pos, char *path)
{
	char buf[256];
	ZPACK_CacheFile *ctmp;
	char *s, *t;
	int i, j, k, l;

	i=pos;
	while(1)
	{
		if(!ctx->root[i].name[0])
		{
			i=ZPACK_GetU32(ctx->root[i].chain);
			if(!i)break;
			continue;
		}

		strcpy(buf, ctx->root[i].name);
		if(*path)sprintf(buf, "%s/%s", path, ctx->root[i].name);

		j=ZPACK_GetU16(ctx->root[i].flags);
		if(j&2)
		{
			j=ZPACK_GetU32(ctx->root[i].crc32);
			if(j)ZPACK_ProcessDir_R(ctx, j, buf);

			i=ZPACK_GetU32(ctx->root[i].chain);
			if(!i)break;
			continue;
		}

//		printf("Process %s\n", buf);

		l=ZPACK_GetU64(ctx->root[i].offset);
		k=ZPACK_GetU32(ctx->root[i].csize);
		ZPACK_InsertSpan(ctx, l, k, ctx->root[i].name);

		if(j&1)
		{
//			printf("Process Frag %s\n", buf);
			ctmp=ZPACK_GetCacheFragmented(ctx, buf);
			if(!ctmp)printf("Fail Process Frag %s\n", buf);
		}

		i=ZPACK_GetU32(ctx->root[i].chain);
		if(!i)break;
	}
}

ZPACK_Context *ZPACK_OpenImage(char *name)
{
	struct ZPACK_Header_ds hdr;
	FILE *fd;
	ZPACK_Context *ctx;
	ulong li;
	int i, j, k;

	fd=fopen(name, "r+b");
	if(!fd)return(NULL);

	fread(&hdr, 1, 16, fd);

	if(ZPACK_GetU32(hdr.magic)!=FCC_ZPAK)
		return(NULL);

	ctx=malloc(sizeof(ZPACK_Context));
	memset(ctx, 0, sizeof(ZPACK_Context));

	ctx->fd=fd;
	ctx->ents=ZPACK_GetU32(hdr.ents);
	ctx->offs=ZPACK_GetU64(hdr.offs);

	fseek(fd, 0, 2);
	ctx->eof=ftell(fd);

	ctx->ients=ctx->ents;

	ctx->root=malloc(ctx->ents*64);
	fseek(fd, ctx->offs, 0);
	fread(ctx->root, 64, ctx->ents, fd);

	//spans for header and root
	ZPACK_InsertSpan(ctx, 0, 64, "$head");
	ZPACK_InsertSpan(ctx, ctx->offs, ctx->ents*64, "$root");

	//process root to rebuild spans and such
	ZPACK_ProcessDir_R(ctx, 0, "");

#if 0
	//spans for each file
	for(i=0; i<ctx->ents; i++)
	{
		li=ZPACK_GetU64(ctx->root[i].offset);
		j=ZPACK_GetU32(ctx->root[i].csize);
		ZPACK_InsertSpan(ctx, li, j, ctx->root[i].name);
	}
#endif

	return(ctx);
}

int ZPACK_CommitRoot(ZPACK_Context *ctx)
{
	int i, j;

//	if(ctx->ents!=ctx->ients)
	if(!ZPACK_CheckSpan(ctx, ctx->offs, ctx->ents*64))
	{
		ctx->ients=ctx->ents;
		j=ctx->ents*64;

//		ctx->offs=ctx->eof;
//		ctx->eof+=j;

		ZPACK_DeleteSpan(ctx, ctx->offs);
		ctx->offs=ZPACK_FindMinSpan(ctx, j);
		if((ctx->offs+j)>ctx->eof)ctx->eof=ctx->offs+j;

		ZPACK_InsertSpan(ctx, ctx->offs, j, "<root>");
	}

	j=ctx->ents*64;
	fseek(ctx->fd, 0, 2);
	i=ftell(ctx->fd);
	if(i<ctx->offs)ZPACK_WritePad(ctx->fd, i-ctx->offs);
	fseek(ctx->fd, ctx->offs, 0);
	fwrite(ctx->root, 1, j, ctx->fd);
}

int ZPACK_CloseImage(ZPACK_Context *ctx)
{
	ZPACK_CommitCacheAll(ctx);
	ZPACK_CommitRoot(ctx);
	fclose(ctx->fd);
}

int ZPACK_FindFreeEntry(ZPACK_Context *ctx, int last)
{
	int i, j;

	for(i=last; i<ctx->ents; i++)
	{
		if(!ctx->root[i].name[0])
			return(i);
	}
	for(i=1; i<last; i++)
	{
		if(!ctx->root[i].name[0])
			return(i);
	}

	i=ctx->ents;
	ctx->ents=(ctx->ents+96)&(~63);
	ctx->root=realloc(ctx->root, ctx->ents*64);
	memset(ctx->root+i, 0, (ctx->ents-i)*64);
	return(i);
}

char *ZPACK_ParseName(char *s, char *obuf)
{
	static char *b36="0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ";
	char buf[256];
	char *t, *s2, *t2;
	int i, h;

	t=buf;
	while(*s && (*s!='/'))*t++=*s++;
	*t++=0;
//	while(*s && (*s=='/'))s++;

	if(strlen(buf)<32)
	{
		strcpy(obuf, buf);
		return(s);
	}

	//name exceeds limit, so mangle
	s2=buf; h=0;
	while(*s2)h=(h>>20)^(h<<3)^(*s2++);
	h&=0xFFFFF;

	s2=buf+strlen(buf);
	while((s2>buf) && (*s2!='.'))s2--;
	if((s2<=buf) || (strlen(s2)>8))s2=NULL;

	i=32-6;
	if(s2)i-=strlen(s2);
	t=buf; t2=obuf;
	while(i--)*t2++=*t++;

	*t2++='~';
	*t2++=b36[(h/46656)%36];
	*t2++=b36[(h/1296)%36];
	*t2++=b36[(h/36)%36];
	*t2++=b36[h%36];

	if(s2)strcpy(t2, s2);

	return(s);
}

int ZPACK_LocateFile_R(ZPACK_Context *ctx, int pos, char *name)
{
	char buf[32];
	char *s, *t;
	int i, j;

//	printf("LookupR %s\n", name);

//	s=name; t=buf;
//	while(*s && (*s!='/'))*t++=*s++;
//	*t++=0;
//	while(*s && (*s=='/'))s++;

	s=ZPACK_ParseName(name, buf);
	while(*s && (*s=='/'))s++;

	i=pos;
	while(1)
	{
//		printf("LStep %d: %s\n", i, ctx->root[i].name);

		if(strcmp(ctx->root[i].name, buf))
		{
			i=ZPACK_GetU32(ctx->root[i].chain);
			if(!i) { i=-1; break; }
			continue;
		}
		break;
	}

	if(i<0)return(i);	//not found

//	printf("Check %d: %s\n", i, ctx->root[i].name);

	if(!(*s))return(i);	//found

	j=ZPACK_GetU16(ctx->root[i].flags);
	if(!(j&2))return(-1);

	j=ZPACK_GetU32(ctx->root[i].crc32);
	if(!j)return(-1);	//dir is empty

	j=ZPACK_LocateFile_R(ctx, j, s);
	return(j);
}

int ZPACK_LocateFile(ZPACK_Context *ctx, char *name)
{
	char *s;
	int i;

//	printf("Locate %s\n", name);

	s=name;
	while(*s && (*s=='/'))s++;
	if(!(*s))return(-1);
	i=ZPACK_LocateFile_R(ctx, 0, s);
	return(i);
}

int ZPACK_CreateFile_R(ZPACK_Context *ctx, int pos, char *name)
{
	char buf[32];
	char *s, *t;
	int i, j, k, sd;

//	printf("CreateR %s\n", name);

//	s=name; t=buf;
//	while(*s && (*s!='/'))*t++=*s++;
//	*t++=0;

	s=ZPACK_ParseName(name, buf);

	sd=(*s=='/')?1:0;
	while(*s && (*s=='/'))s++;

	i=pos; j=0;
	while(1)
	{
//		printf("Step %d: %s\n", i, ctx->root[i].name);
		if(strcmp(ctx->root[i].name, buf))
		{
			j=ZPACK_GetU32(ctx->root[i].chain);
			if(!j) { j=-1; break; }
			i=j;
			continue;
		}
		break;
	}

	if(j<0)	//not found, create
	{
//		printf("Create %s\n", buf);

		j=ZPACK_FindFreeEntry(ctx, i?i:1);
		memset(&ctx->root[j], 0, 64);
		strcpy(ctx->root[j].name, buf);

		if(i!=j)ZPACK_SetU32(ctx->root[i].chain, j);

		if(*s || sd)
		{
			ZPACK_SetU16(ctx->root[j].flags, 2);
//			k=ZPACK_FindFreeEntry(ctx, j);
//			ZPACK_SetU32(ctx->root[j].crc32, k);
		}
		if(*s)
		{
			k=ZPACK_FindFreeEntry(ctx, j);
			ZPACK_SetU32(ctx->root[j].crc32, k);

			k=ZPACK_CreateFile_R(ctx, k, s);
			return(k);
		}

//		printf("Create %s: OK\n", buf);

		return(j);
	}

	if(!(*s))	//found, pre-existing
	{
		//need to destroy any caches and truncate file

		k=ZPACK_GetU64(ctx->root[i].offset);
		ZPACK_DeleteSpan(ctx, k);

		ZPACK_SetU16(ctx->root[i].flags, 0);
		ZPACK_SetU16(ctx->root[i].method, 0);
		ZPACK_SetU32(ctx->root[i].crc32, 0);
		ZPACK_SetU32(ctx->root[i].usize, 0);
		ZPACK_SetU32(ctx->root[i].csize, 0);
		ZPACK_SetU64(ctx->root[i].offset, 0);

		return(i);
	}

	//something was found, verify if it is a directory
	j=ZPACK_GetU16(ctx->root[i].flags);
	if(!(j&2))return(-1);

	//sub directory
	j=ZPACK_GetU32(ctx->root[i].crc32);
	if(!j)
	{
		//dir empty, create contents
		j=ZPACK_FindFreeEntry(ctx, i);
		ZPACK_SetU32(ctx->root[i].crc32, j);
	}

	j=ZPACK_CreateFile_R(ctx, j, s);
	return(j);
}

int ZPACK_CreateFile(ZPACK_Context *ctx, char *name)
{
	char *s;
	int i;

	s=name;
	while(*s && (*s=='/'))s++;
	if(!(*s))return(-1);
	i=ZPACK_CreateFile_R(ctx, 0, s);
	return(i);
}

int ZPACK_DeleteFile_R(ZPACK_Context *ctx, int pos, int up, char *name)
{
	char buf[32];
	char *s, *t;
	int i, j, k;

	s=ZPACK_ParseName(name, buf);
	while(*s && (*s=='/'))s++;

	i=pos; j=-1;
	while(1)
	{
		if(strcmp(ctx->root[i].name, buf))
		{
			j=i;
			i=ZPACK_GetU32(ctx->root[i].chain);
			if(!i) { i=-1; break; }
			continue;
		}
		break;
	}

	if(i<0)return(-1);	//not found

	if(!(*s))	//found
	{
		k=ZPACK_GetU64(ctx->root[i].offset);
		ZPACK_DeleteSpan(ctx, k);

		if(j>=0)
		{
			//not first item, so unlink
			k=ZPACK_GetU32(ctx->root[i].chain);
			ZPACK_SetU32(ctx->root[j].chain, k);
			memset(&ctx->root[i], 0, 64);
		}else if(up>=0)
		{
			//first item, but a parent exists
			k=ZPACK_GetU32(ctx->root[i].chain);
			ZPACK_SetU32(ctx->root[up].crc32, k);
			memset(&ctx->root[i], 0, 64);
		}else
		{
			//first entry of root, be more careful
			memset(&ctx->root[i].name, 0, 32);

			ZPACK_SetU16(ctx->root[i].flags, 0);
			ZPACK_SetU16(ctx->root[i].method, 0);
			ZPACK_SetU32(ctx->root[i].crc32, 0);
			ZPACK_SetU32(ctx->root[i].usize, 0);
			ZPACK_SetU32(ctx->root[i].csize, 0);
			ZPACK_SetU64(ctx->root[i].offset, 0);
		}

		return(i);
	}

	//something was found, verify if it is a directory
	j=ZPACK_GetU16(ctx->root[i].flags);
	if(!(j&2))return(-1);

	//step into sub-directory
	j=ZPACK_GetU32(ctx->root[i].crc32);
	if(!j)return(-1);

	j=ZPACK_DeleteFile_R(ctx, j, i, s);
	return(j);
}

int ZPACK_DeleteFile(ZPACK_Context *ctx, char *name)
{
	char *s;
	int i;

	s=name;
	while(*s && (*s=='/'))s++;
	if(!(*s))return(-1);

	ZPACK_DropCache(ctx, s);

	i=ZPACK_DeleteFile_R(ctx, 0, -1, s);
	return(i);
}


int ZPACK_StatFile(ZPACK_Context *ctx, char *name, ZPACK_STAT *stat)
{
	ZPACK_CacheFile *cur;
	int i, j;

	i=ZPACK_LocateFile(ctx, name);
	if(i<0)return(-1);

	stat->date_time=ZPACK_GetU32(ctx->root[i].date_time);
	stat->flags=ZPACK_GetU16(ctx->root[i].flags);
	stat->method=ZPACK_GetU16(ctx->root[i].method);
	stat->crc32=ZPACK_GetU32(ctx->root[i].crc32);
	stat->usize=ZPACK_GetU32(ctx->root[i].usize);
	stat->csize=ZPACK_GetU32(ctx->root[i].csize);
	stat->offset=ZPACK_GetU64(ctx->root[i].offset);

	if(stat->flags&1)
	{
		stat->usize=0;
		stat->csize=0;

		cur=ZPACK_GetCacheFragmented(ctx, name);
		if(cur)
		{
			stat->usize=cur->lsize;
			stat->csize=0;
			stat->crc32=(uint)-1;

			for(i=0; i<cur->n_frag; i++)
				stat->csize+=cur->frag[i].size;
			for(i=0; i<cur->n_frag; i++)
			{
				if((i+1)==cur->n_frag)
				{
					j=cur->lsize&((1<<cur->frag_bits)-1);
					if(cur->lsize && !j)
						j=1<<cur->frag_bits;
				}else j=1<<cur->frag_bits;

				stat->crc32=ZPACK_CombineCRC32(
					stat->crc32, cur->frag[i].crc,
					j);
			}
		}
	}

	return(0);
}

int ZPACK_SetFileTime(ZPACK_Context *ctx, char *name, uint t)
{
	int i;

	i=ZPACK_LocateFile(ctx, name);
	if(i<0)return(-1);
	ZPACK_SetU32(ctx->root[i].date_time, t);
	return(0);
}
