ZPACK_FILE *ZPACK_OpenFile(ZPACK_Context *ctx, char *name, char *mode)
{
	ZPACK_CacheFile *cf;
	ZPACK_FILE *tmp;
	int i;

	while(*name && (*name=='/'))name++;
	if(!(*name))return(NULL);

	if(*mode=='r')
	{
		cf=ZPACK_GetCache(ctx, name);
		if(!cf)return(NULL);

		tmp=malloc(sizeof(ZPACK_FILE));
		memset(tmp, 0, sizeof(ZPACK_FILE));

		tmp->ctx=ctx;
		tmp->cache=cf;
		tmp->offs=0;

		return(tmp);
	}

	if(*mode=='w')
	{
		ZPACK_DropCache(ctx, name);
		i=ZPACK_CreateFile(ctx, name);
		if(i<0)
		{
			printf("Can't create entry for %s\n", name);
			return(NULL);
		}

		cf=ZPACK_GetCache(ctx, name);
		if(!cf)
		{
			printf("Can't get cache for %s\n", name);
			return(NULL);
		}

		tmp=malloc(sizeof(ZPACK_FILE));
		memset(tmp, 0, sizeof(ZPACK_FILE));

		tmp->ctx=ctx;
		tmp->cache=cf;
		tmp->offs=0;

		return(tmp);
	}

	if(*mode=='a')
	{
		cf=ZPACK_GetCache(ctx, name);
		if(!cf)return(NULL);

		tmp=malloc(sizeof(ZPACK_FILE));
		memset(tmp, 0, sizeof(ZPACK_FILE));

		tmp->ctx=ctx;
		tmp->cache=cf;
		tmp->offs=cf->lsize;

		return(tmp);
	}

	return(NULL);
}

int ZPACK_Close(ZPACK_FILE *fd)
{
	ZPACK_CommitCache(fd->ctx, fd->cache);
	ZPACK_CommitRoot(fd->ctx);
	free(fd);
}

int ZPACK_Read(ZPACK_FILE *fd, byte *buf, int sz)
{
	int i, j;

	ZPACK_MoveFrontMRU(fd->ctx, fd->cache);

	i=fd->cache->lsize-fd->offs;
	if(sz<i)i=sz;
	memcpy(buf, fd->cache->buffer+fd->offs, i);
	fd->offs+=i;
	return(i);
}

int ZPACK_Write(ZPACK_FILE *fd, byte *buf, int sz)
{
	int i, j, k;

	ZPACK_MoveFrontMRU(fd->ctx, fd->cache);
	fd->cache->dirty=1;

	//need to expand buffer
	if((fd->offs+sz)>=fd->cache->bsize)
	{
		i=fd->cache->lsize;
		j=(fd->offs+sz+6144)&(~4095);
		fd->cache->bsize=j;

		fd->cache->buffer=realloc(fd->cache->buffer, j);
		memset(fd->cache->buffer+i, 0, j-i);
	}

	if((fd->offs+sz)>fd->cache->lsize)
	{
		fd->cache->lsize=fd->offs+sz;

		if((fd->cache->lsize>=65536) && !fd->cache->frag)
		{
			fd->cache->m_frag=16;
			fd->cache->n_frag=(fd->cache->lsize+65535)/65536;
			fd->cache->frag_bits=16;

			fd->cache->frag=malloc(fd->cache->m_frag*
				sizeof(ZPACK_FragSpan));
			memset(fd->cache->frag, 0, fd->cache->m_frag*
				sizeof(ZPACK_FragSpan));

			for(i=0; i<fd->cache->n_frag; i++)
				fd->cache->frag[i].flags=1;
		}

		if(fd->cache->frag)
		{
			k=fd->cache->frag_bits;

			i=(fd->cache->lsize+((1<<k)-1))>>k;
			if(i>fd->cache->m_frag)
			{
				fd->cache->m_frag+=fd->cache->m_frag>>1;
				fd->cache->frag=realloc(fd->cache->frag,
					fd->cache->m_frag*
					sizeof(ZPACK_FragSpan));
			}
			if(i>fd->cache->n_frag)
			{
				for(j=fd->cache->n_frag; j<i; j++)
				{
					memset(&fd->cache->frag[j], 0,
						sizeof(ZPACK_FragSpan));
					fd->cache->frag[j].flags=1;
				}
				fd->cache->n_frag=i;
			}
		}
	}

	if(fd->cache->frag)
	{
		k=fd->cache->frag_bits;
		i=fd->cache->offs>>k;
		j=(sz+((1<<k)-1))>>k;
		while(j--)fd->cache->frag[i++].flags|=1;
	}

	memcpy(fd->cache->buffer+fd->offs, buf, sz);
	fd->offs+=sz;
	return(sz);
}

uint ZPACK_Seek(ZPACK_FILE *fd, uint offs, int rel)
{
	switch(rel)
	{
	case 0:
		fd->offs=offs;
		break;
	case 1:
		fd->offs+=offs;
		break;
	case 2:
		fd->offs=fd->cache->lsize+offs;
		break;
	default:
		break;
	}

	return(fd->offs);
}

uint ZPACK_Tell(ZPACK_FILE *fd)
{
	return(fd->offs);
}
